(* $Id: hydro_builtin_util.ml 15317 2007-11-22 17:15:35Z gerd $ *)

open Hydro_types

let system =
  Hydro_lm.create_system()

let () =
  Hydro_builtin.fill_system system

let derive_proxy_env resolver (pe : Hydro_proxy.proxy_env_t)  =
  (* Derive a proxy_env from pe that shares the pool and the esys *)
  ( object
      method event_system = pe # event_system
      method system = system
      method proxy_resolver = resolver
      method client_pool = pe # client_pool
      method default_proxy_conf = Hydro_proxy.proxy_conf()
    end : Hydro_proxy.proxy_env_t
  )


let esc_re = 
  Pcre.regexp "[/\\x5c\\x00-\\x1f\\x7f-\\xff]"

let escape_identity_string s =
  Pcre.substitute
    ~rex:esc_re
    ~subst:(fun m ->
	      match m with
		| "/" -> "\\/"
		| "\\" -> "\\\\"
		| "'" -> "\\'"
		| "\"" -> "\\\""
		| "\008" -> "\\b"
		| "\012" -> "\\f"
		| "\010" -> "\\n"
		| "\013" -> "\\r"
		| "\009" -> "\\t"
		| _ -> Printf.sprintf "\\%03o" (Char.code m.[0])
	   )
    s

let string_of_identity (id : identity) =
    if id#category = "" then
      escape_identity_string id#name
    else
      escape_identity_string id#category ^ "/" ^ escape_identity_string id#name

let unesc_re =
  Pcre.regexp "\\\\(?:/|'|\"|b|f|n|r|t|\\\\|[0-3][0-7][0-7]|[0-7][0-7]|[0-7]|.)"

type parsed =
  | Literal of string
  | Slash

let rec split_at_slash toks =
  match toks with
    | Slash :: toks' ->
	( [], toks' )
    | tok :: toks' ->
	let (l,r) = split_at_slash toks' in
	(tok :: l, r)
    | [] ->
	raise Not_found
	

let concat_toks toks =
  let b = Buffer.create 100 in
  List.iter
    (function 
       | Slash -> Buffer.add_char b '/'
       | Literal s -> Buffer.add_string b s
    )
    toks;
  Buffer.contents b


let identity_of_string s =
  let parsed =
    List.flatten
      (List.map
	 (function
	    | Pcre.Text t ->
		(* Look for "/" *)
		( try
		    let k = String.index t '/' in
		    [ Literal(String.sub t 0 k);
		      Slash;
		      Literal(String.sub t (k+1) (String.length t - k - 1))
		    ]
		  with
		    | Not_found -> [ Literal t ]
		)
	    | Pcre.Delim d ->
		( match d.[1] with
		    | '/' | '\'' | '"' | '\\' ->
			[ Literal (String.sub d 1 1) ]
		    | 'n' ->
			[ Literal "\010" ]
		    | 'b' ->
			[ Literal "\008" ]
		    | 'f' ->
			[ Literal "\012" ]
		    | 'r' ->
			[ Literal "\013" ]
		    | 't' ->
			[ Literal "\009" ]
		    | '0'..'7' ->
			[ Literal
			    (String.make 1
			       (Char.chr
				  (int_of_string
				     ("0o" ^ 
					String.sub d 1 (String.length d - 1))))) 
			]
		    | _ ->
			[ Literal (String.sub d 1 1) ]
		)
	    | _ -> assert false
	 )
	 (Pcre.full_split
	    ~rex:unesc_re
	    s
	 )
      ) in
  try
    let (before_slash, after_slash) = split_at_slash parsed in
    let name = concat_toks after_slash in
    let category = concat_toks before_slash in
    if name = "" && category = "" then
      failwith "Illegal identity string";
    ( object
	method name = name
	method category = category
      end : identity
    )
  with
    | Not_found ->
	let name = concat_toks parsed in
	( object
	    method name = name
	    method category = ""
	  end : identity
	)
