(* $Id: hydro_util.ml 16055 2008-01-11 16:04:07Z gerd $ *)

open Hydro_types

let t_proxy : htype =
  `Struct [| "name",     `String;
	     "category", `String;
	     "facet",    (`Sequence `String);
	     "mode",     `Byte;
	     "secure",   `Bool
	  |]

let t_proxy_alt : htype =
  `Struct [| "facet",    (`Sequence `String);
	     "mode",     `Byte;
	     "secure",   `Bool
	  |]


let t_tcp_endpoint : htype =
  `Struct [| "host",     `String;
	     "port",     `Int;
	     "timeout",  `Int32;
	     "compress", `Bool
	  |]


let t_udp_endpoint : htype =
  `Struct [| "host",         `String;
	     "port",        `Int;
	     "proto_major", `Int;
	     "proto_minor", `Int;
	     "enc_major",   `Int;
	     "enc_minor",   `Int;
	     "compress",    `Bool
	  |]


let t_msg_request : htype =
  `Struct [| "requestId",   `Int32;
	     "id.name",     `String;
	     "id.category", `String;
	     "facet",       (`Sequence `String);
	     "operation",   `String;
	     "mode",        `Byte;
	     "context",     (`Dictionary(`String, `String));
	  |]


let t_msg_batch_request : htype =
  `Struct [| "id.name",     `String;
	     "id.category", `String;
	     "facet",       (`Sequence `String);
	     "operation",   `String;
	     "mode",        `Byte;
	     "context",     (`Dictionary(`String, `String));
	  |]

let t_msg_reply : htype =
  `Struct [| "requestId",   `Int32;
	     "replyStatus", `Byte;
	  |]

let t_msg_reply_error : htype =
  `Struct [| "id.name",     `String;
	     "id.category", `String;
	     "facet",       (`Sequence `String);
	     "operation",   `String;
	  |]

let endpoint_type =
  function
    | `TCP _ -> `TCP
    | `UDP _ -> `UDP
    | `SSL _ -> `SSL
    | `Unknown(n,_) -> `Unknown n


let string_of_endpoint_type =
  function
    | `TCP -> "TCP"
    | `UDP -> "UDP"
    | `SSL -> "SSL"
    | `Unknown n -> string_of_int n

let endpoint_cmp (ep1:endpoint) (ep2:endpoint) =
  match ep1 with
    | `TCP tcp1 ->
	( match ep2 with
	    | `TCP tcp2 ->
		Pervasives.compare
		  (tcp1#host, tcp1#port, tcp1#timeout, tcp1#compress)
		  (tcp2#host, tcp2#port, tcp2#timeout, tcp2#compress)
	    | _ ->
		(-1)
	)
    | `UDP udp1 ->
	( match ep2 with
	    | `TCP _ ->
		1
	    | `UDP udp2 ->
		Pervasives.compare
		  (udp1#host, udp1#port, udp1#proto_major, udp1#proto_minor,
		   udp1#enc_major, udp1#enc_minor, udp1#compress)
		  (udp2#host, udp2#port, udp2#proto_major, udp2#proto_minor,
		   udp2#enc_major, udp2#enc_minor, udp2#compress)
	    | _ ->
		(-1)
	)
    | `SSL ssl1 ->
	( match ep2 with
	    | `TCP _ | `UDP _ ->
		1
	    | `SSL ssl2 ->
		Pervasives.compare
		  (ssl1#host, ssl1#port, ssl1#timeout, ssl1#compress)
		  (ssl2#host, ssl2#port, ssl2#timeout, ssl2#compress)
	    | _ ->
		(-1)
	)
    | `Unknown(n1,s1) ->
	( match ep2 with
	    | `TCP _ | `UDP _ | `SSL _ ->
		1
	    | `Unknown(n2,s2) ->
		Pervasives.compare (n1,s1) (n2,s2)
	)

let network_port_of_connector c : network_port option =
  (* as good as possible... we don't do dns lookups, however.
   *)
  match c with
    | `Endpoint(_,port_opt) ->
	port_opt

let host_port_of_endpoint ep =
  match ep with
    | `TCP s | `SSL s ->
	Some(s#host, s#port)
    | `UDP s ->
	Some(s#host, s#port)
    | _ ->
	None


let connector_of_resolved_endpoint ep addr =
  match ep with
    | `TCP s | `SSL s ->
	`Endpoint(ep, Some(`TCP(addr, s#port)))
    | `UDP s ->
	`Endpoint(ep, Some(`UDP(addr, s#port)))
    | _ ->
	`Endpoint(ep, None)


(* for debugging: *)

let pv_to_string : protocol_violation -> string =
  function
    | `MessageFormatViolation s ->
	"MessageFormatViolation(" ^ s ^ ")"
    | `CompressionNotSupported ->
	"CompressionNotSupported"
    | `BadMessageType ->
	"BadMessageType"

let lim_to_string : limitation -> string =
  function
    | `UnsupportedEncodingVersion ->
	"UnsupportedEncodingVersion"
    | `UnsupportedProtocolVersion ->
	"UnsupportedProtocolVersion"
    | `UnsupportedEndpointType n ->
	"UnsupportedEndpointType(" ^ string_of_int n ^ ")"

let rec exn_to_string : exn -> string =
  function
    | Protocol_violation pv ->
	"Protocol_violation(" ^ pv_to_string pv ^ ")"
    | Marshal_error s ->
	"Marshal_error(" ^ s ^ ")"
    | Unmarshal_error s ->
	"Unmarshal_error(" ^ s ^ ")"
    | Limitation lim ->
	"Limitation(" ^ lim_to_string lim ^ ")"
    | Client_condition cc ->
	"Client_condition(" ^ cc_to_string cc ^ ")"
    | Proxy_error pe ->
	"Proxy_error(" ^ pe_to_string pe ^ ")"
    | Domain_not_found name ->
	"Domain_not_found(" ^ name ^ ")"
    (* Because of buggy Printexc.to_string: *)
    | Invalid_argument msg ->
	"Invalid_argument(" ^ msg ^ ")"
    | Failure msg ->
	"Failure(" ^ msg ^ ")"
    | other ->
	Printexc.to_string other

and pe_to_string =
  function
    | `NoCallableEndpointFound -> "NoCallableEndpointFound"
    | `NoEndpointIsReachable -> "NoEndpointIsReachable"
    | `NoLocatorIsReachable -> "NoLocatorIsReachable"
    | `ProxyIsDown -> "ProxyIsDown"

and cc_to_string =
  function
    | `Message_lost flag -> "Message_lost(" ^ string_of_bool flag ^ ")"
    | `Message_timeout -> "Message_timeout"
    | `Transport_timeout -> "Transport_timeout"
    | `Connect_timeout -> "Connect_timeout"
    | `Connect_error x -> "Connect_error(" ^ exn_to_string x ^ ")"
    | `Client_is_down -> "Client_is_down"
    | `User_exception _ -> "User_exception"
    | `Object_does_not_exist _ -> "Object_does_not_exist"
    | `Facet_does_not_exist _ -> "Facet_does_not_exist"
    | `Operation_does_not_exist _ -> "Operation_does_not_exist"
    | `Unknown_local_exception s -> "Unknown_local_exception(" ^ s ^ ")"
    | `Unknown_user_exception s -> "Unknown_user_exception(" ^ s ^ ")"
    | `Unknown_exception s -> "Unknown_exception(" ^ s ^ ")"
    | `Error x -> "Error(" ^ exn_to_string x ^ ")"

let result_to_string =
  function
    | `Success eb ->
	"Success(" ^ string_of_int eb.encap_len ^ ")"
    | `User_exception eb ->
	"User_exception(" ^ string_of_int eb.encap_len ^ ")"
    | `Object_does_not_exist _ ->
	"Object_does_not_exist"
    | `Facet_does_not_exist _ ->
	"Facet_does_not_exist"
    | `Operation_does_not_exist _ ->
	"Operation_does_not_exist"
    | `Unknown_local_exception s ->
	"Unknown_local_exception(" ^ s ^ ")"
    | `Unknown_user_exception s ->
	"Unknown_user_exception(" ^ s ^ ")"
    | `Unknown_exception s ->
	"Unknown_exception(" ^ s ^ ")"


let rec htype_to_string =
  function
    | `Void -> "void"
    | `Bool -> "bool"
    | `Byte -> "byte"
    | `Short -> "short"
    | `Int -> "int"
    | `Int32 -> "int32"
    | `Long -> "long"
    | `Float -> "float"
    | `Double -> "double"
    | `String -> "string"
    | `Enum _ -> "enum"
    | `Struct s ->
	"{" ^
	  (String.concat "; " 
	     (List.map
		(fun (n,t) -> n ^ "=" ^ htype_to_string t)
		(Array.to_list s))) ^
	  "}"
    | `Sequence s ->
	htype_to_string s ^ " sequence";
    | `Dictionary (s,t) ->
	"(" ^ htype_to_string s ^ "," ^  htype_to_string t ^ ") dictionary"
    | `Proxy p ->
	p ^ "*"
    | `Class c ->
	c ^ "@"
